"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.

The dialog can be seen as a decision tree, where based on user input new questions (nodes in the tree)
are asked. This tree is realized by recursive function calls, each function call representing a node
that modifies the user object based and returns a call to the next configuration step based on the
user's configuration.
"""
from __future__ import annotations, print_function, unicode_literals

import enum
import os
import pathlib
from dataclasses import dataclass
from os.path import expanduser as user
from typing import Dict, List, Optional, Union

import cowsay
from kubernetes import client, config
from pyfiglet import Figlet
from yaspin import yaspin

from now.deployment.deployment import cmd
from now.thridparty.PyInquirer import Separator
from now.thridparty.PyInquirer.prompt import prompt
from now.utils import sigmap

cur_dir = pathlib.Path(__file__).parent.resolve()
NEW_CLUSTER = {'name': '🐣 create new', 'value': 'new'}
AVAILABLE_SOON = 'will be available in upcoming versions'
QUALITY_MAP = {
    'medium': ('ViT-B32', 'openai/clip-vit-base-patch32'),
    'good': ('ViT-B16', 'openai/clip-vit-base-patch16'),
    'excellent': ('ViT-L14', 'openai/clip-vit-large-patch14'),
}
AVAILABLE_DATASET = {
    'image': [
        'best-artworks',
        'nft-monkey',
        'tll',
        'bird-species',
        'stanford-cars',
        'deepfashion',
        'nih-chest-xrays',
        'geolocation-geoguessr',
    ],
    'music': [
        'music-genres-small',
        'music-genres-large',
    ],
    'text': ['rock-lyrics', 'pop-lyrics', 'rap-lyrics', 'indie-lyrics', 'metal-lyrics'],
}


class Modalities(str, enum.Enum):
    IMAGE = 'image'
    MUSIC = 'music'
    TEXT = 'text'


@dataclass
class UserInput:
    output_modality: Optional[Modalities] = None

    # data related
    data: Optional[str] = None
    is_custom_dataset: Optional[bool] = None

    custom_dataset_type: Optional[str] = None
    dataset_secret: Optional[str] = None
    dataset_url: Optional[str] = None
    dataset_path: Optional[str] = None

    # model related
    quality: Optional[str] = None
    model_variant: Optional[str] = None

    # cluster related
    cluster: Optional[str] = None
    create_new_cluster: Optional[bool] = None
    new_cluster_type: Optional[str] = None


def configure_user_input(**kwargs) -> UserInput:
    print_headline()
    user_input = _configure_output_modality(UserInput(), **kwargs)
    return user_input


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and kwargs.get(attribute):
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        if attribute in answer:
            return answer[attribute]
        else:
            print("\n" * 10)
            cowsay.cow('see you soon 👋')
            exit(0)


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina now trains a model, pushes it to the jina hub'
        ', deploys a flow and a frontend app in the cloud or locally. \nCheckout one of the demo cases or bring '
        'your own data.\n'
    )
    print(
        'If you want learn more about our framework please visit: https://docs.jina.ai/'
    )
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def _configure_output_modality(user_input: UserInput, **kwargs) -> UserInput:
    modality = _prompt_value(
        name='output_modality',
        choices=[
            {'name': '🏞 Image Search', 'value': Modalities.IMAGE},
            {'name': '📝 Text Search (experimental)', 'value': Modalities.TEXT},
            {
                'name': '🥁 Music Search',
                'value': Modalities.MUSIC,
                'disabled': AVAILABLE_SOON,
            },
        ],
        prompt_message='Which modalities you want to work with?',
        prompt_type='list',
        **kwargs,
    )
    user_input.output_modality = modality
    if modality == Modalities.IMAGE:
        return _configure_dataset_image(user_input, **kwargs)
    elif modality == Modalities.TEXT:
        return _configure_dataset_text(user_input, **kwargs)
    else:
        return _configure_dataset_music(user_input, **kwargs)


def _configure_dataset_image(user_input: UserInput, **kwargs) -> UserInput:
    dataset = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🖼  artworks (≈8K docs)', 'value': 'best-artworks'},
            {
                'name': '💰 nft - bored apes (10K docs)',
                'value': 'nft-monkey',
            },
            {'name': '👬 totally looks like (≈12K docs)', 'value': 'tll'},
            {'name': '🦆 birds (≈12K docs)', 'value': 'bird-species'},
            {'name': '🚗 cars (≈16K docs)', 'value': 'stanford-cars'},
            {
                'name': '🏞 geolocation (≈50K docs)',
                'value': 'geolocation-geoguessr',
            },
            {'name': '👕 fashion (≈53K docs)', 'value': 'deepfashion'},
            {
                'name': '☢️ chest x-ray (≈100K docs)',
                'value': 'nih-chest-xrays',
            },
            Separator(),
            {
                'name': '✨ custom',
                'value': 'custom',
            },
        ],
        **kwargs,
    )
    user_input.data = dataset
    return _configure_dataset(user_input, **kwargs)


def _configure_dataset_text(user_input: UserInput, **kwargs) -> UserInput:
    dataset = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🎤 rock lyrics (200K docs)', 'value': 'rock-lyrics'},
            {'name': '🎤 pop lyrics (200K docs)', 'value': 'pop-lyrics'},
            {'name': '🎤 rap lyrics (200K docs)', 'value': 'rap-lyrics'},
            {'name': '🎤 indie lyrics (200K docs)', 'value': 'indie-lyrics'},
            {'name': '🎤 metal lyrics (200K docs)', 'value': 'metal-lyrics'},
        ],
        **kwargs,
    )
    user_input.data = dataset
    return _configure_dataset(user_input, **kwargs)


def _configure_dataset_music(user_input: UserInput, **kwargs):
    dataset = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🎸 music small (≈2K docs)', 'value': 'music-genres-small'},
            {'name': '🎸 music large (≈10K docs)', 'value': 'music-genres-large'},
            Separator(),
            {
                'name': '✨ custom',
                'value': 'custom',
            },
        ],
        **kwargs,
    )
    user_input.data = dataset
    return _configure_dataset(user_input, **kwargs)


def _configure_dataset(user_input: UserInput, **kwargs) -> UserInput:
    dataset = user_input.data
    if dataset in AVAILABLE_DATASET[user_input.output_modality]:
        user_input.is_custom_dataset = False
        if user_input.output_modality == Modalities.MUSIC:
            return _configure_cluster(user_input, **kwargs)
        else:
            return _configure_quality(user_input, **kwargs)
    else:
        user_input.is_custom_dataset = True
        if dataset == 'custom':
            return _configure_custom_dataset(user_input, **kwargs)
        else:
            _parse_custom_data_from_cli(dataset, user_input)
            if user_input.output_modality == Modalities.MUSIC:
                return _configure_cluster(user_input, **kwargs)
            else:
                return _configure_quality(user_input, **kwargs)


def _configure_custom_dataset(user_input: UserInput, **kwargs) -> UserInput:
    def configure_docarray() -> UserInput:
        dataset_secret = _prompt_value(
            name='dataset_secret',
            prompt_message='Please enter your docarray secret.',
            prompt_type='password',
        )
        user_input.dataset_secret = dataset_secret
        if user_input.output_modality == Modalities.IMAGE:
            return _configure_quality(user_input, **kwargs)
        else:
            return _configure_cluster(user_input, **kwargs)

    def configure_url() -> UserInput:
        dataset_url = _prompt_value(
            name='dataset_url',
            prompt_message='Please paste in your url for the docarray.',
            prompt_type='input',
        )
        user_input.dataset_url = dataset_url
        if user_input.output_modality == Modalities.IMAGE:
            return _configure_quality(user_input, **kwargs)
        else:
            return _configure_cluster(user_input, **kwargs)

    def configure_path() -> UserInput:
        dataset_path = _prompt_value(
            name='dataset_path',
            prompt_message='Please enter the path to the local folder.',
            prompt_type='input',
        )
        user_input.dataset_path = dataset_path
        if user_input.output_modality == Modalities.IMAGE:
            return _configure_quality(user_input, **kwargs)
        else:
            return _configure_cluster(user_input, **kwargs)

    custom_dataset_type = _prompt_value(
        name='custom_dataset_type',
        prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
        choices=[
            {
                'name': 'docarray.pull id (recommended)',
                'value': 'docarray',
            },
            {
                'name': 'docarray URL',
                'value': 'url',
            },
            {
                'name': 'local mounted path',
                'value': 'path',
            },
        ],
        **kwargs,
    )

    user_input.custom_dataset_type = custom_dataset_type
    if custom_dataset_type == 'docarray':
        return configure_docarray()
    if custom_dataset_type == 'url':
        return configure_url()
    if custom_dataset_type == 'path':
        return configure_path()


def _configure_cluster(user_input: UserInput, **kwargs) -> UserInput:
    def configure_new_cluster() -> UserInput:
        new_cluster_type = _prompt_value(
            name='new_cluster_type',
            choices=[
                {
                    'name': '📍 local (Kubernetes in Docker)',
                    'value': 'local',
                },
                {'name': '⛅️ Google Kubernetes Engine', 'value': 'gke'},
                {
                    'name': '⛅️ Jina - Flow as a Service',
                    'disabled': AVAILABLE_SOON,
                },
                {
                    'name': '⛅️ Amazon Elastic Kubernetes Service',
                    'disabled': AVAILABLE_SOON,
                },
                {
                    'name': '⛅️ Azure Kubernetes Service',
                    'disabled': AVAILABLE_SOON,
                },
                {
                    'name': '⛅️ DigitalOcean Kubernetes',
                    'disabled': AVAILABLE_SOON,
                },
            ],
            prompt_message='Where do you want to create a new cluster?',
            prompt_type='list',
            **kwargs,
        )
        user_input.create_new_cluster = True
        user_input.new_cluster_type = new_cluster_type
        if user_input.new_cluster_type == 'gke':
            _maybe_install_gke(**kwargs)
        return user_input

    choices = _construct_cluster_choices(
        active_context=kwargs.get('active_context'), contexts=kwargs.get('contexts')
    )
    cluster = _prompt_value(
        name='cluster',
        choices=choices,
        prompt_message='Where do you want to deploy your search engine?',
        prompt_type='list',
        **kwargs,
    )
    if cluster == NEW_CLUSTER['value']:
        user_input.cluster = cluster
        return configure_new_cluster()
    else:
        user_input.cluster = cluster
        if not _cluster_running(cluster):
            print(f'Cluster {cluster} is not running. Please select a different one.')
            return _configure_cluster(user_input, **kwargs)
        else:
            return user_input


def _configure_quality(user_input: UserInput, **kwargs) -> UserInput:
    quality = _prompt_value(
        name='quality',
        choices=[
            {'name': '🦊 medium (≈3GB mem, 15q/s)', 'value': 'medium'},
            {'name': '🐻 good (≈3GB mem, 2.5q/s)', 'value': 'good'},
            {
                'name': '🦄 excellent (≈4GB mem, 0.5q/s)',
                'value': 'excellent',
            },
        ],
        prompt_message='What quality do you expect?',
        prompt_type='list',
        **kwargs,
    )
    if quality == 'medium':
        print('  🚀 you trade-off a bit of quality for having the best speed')
    elif quality == 'good':
        print('  ⚖️ you have the best out of speed and quality')
    elif quality == 'excellent':
        print('  ✨ you trade-off speed to having the best quality')

    user_input.quality = quality
    _, user_input.model_variant = QUALITY_MAP[quality]
    return _configure_cluster(user_input, **kwargs)


def _construct_cluster_choices(active_context, contexts):
    context_names = _get_context_names(contexts, active_context)
    choices = [NEW_CLUSTER]
    if len(context_names) > 0 and len(context_names[0]) > 0:
        choices = context_names + choices
    return choices


def _prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
        qs['type'] = 'list'
    return maybe_prompt_user(qs, name, **kwargs)


def _get_context_names(contexts, active_context=None):
    names = [c for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context)
        names = [active_context] + names
    return names


def _cluster_running(cluster):
    config.load_kube_config(context=cluster)
    v1 = client.CoreV1Api()
    try:
        v1.list_namespace()
    except Exception as e:
        return False
    return True


def _maybe_install_gke(os_type: str, arch: str):
    out, _ = cmd('which gcloud')
    if not out:
        if not os.path.exists(user('~/.cache/jina-now/google-cloud-sdk')):
            with yaspin(
                sigmap=sigmap, text='Setting up gcloud', color='green'
            ) as spinner:
                cmd(
                    f'/bin/bash {cur_dir}/scripts/install_gcloud.sh {os_type} {arch}',
                )
                spinner.ok('🛠️')


def _parse_custom_data_from_cli(data: str, user_input: UserInput):
    try:
        data = os.path.expanduser(data)
    except Exception:
        pass
    if os.path.exists(data):
        user_input.custom_dataset_type = 'path'
        user_input.dataset_path = data
    elif 'http' in data:
        user_input.custom_dataset_type = 'url'
        user_input.dataset_url = data
    else:
        user_input.custom_dataset_type = 'docarray'
        user_input.dataset_secret = data



import json
import pathlib
from os.path import expanduser as user

import cowsay
from yaspin import yaspin

from now.deployment.deployment import cmd
from now.dialog import maybe_prompt_user
from now.utils import custom_spinner, sigmap

cur_dir = pathlib.Path(__file__).parent.resolve()


def ask_projects(options):
    options = sorted(options, key=lambda x: x.lower())
    questions = [
        {
            'type': 'list',
            'name': 'project',
            'message': 'What project you want to use for kubernetes deployment?',
            'choices': options,
        }
    ]
    return maybe_prompt_user(questions, 'project')


def ask_regions(options):
    questions = [
        {
            'type': 'list',
            'name': 'region',
            'message': 'Which region to chose?',
            'choices': options,
            'filter': lambda val: val.lower(),
        }
    ]
    return maybe_prompt_user(questions, 'region')


def ask_zones(options):
    questions = [
        {
            'type': 'list',
            'name': 'zone',
            'message': 'Which zone you would like to select?',
            'choices': options,
        }
    ]
    return maybe_prompt_user(questions, 'zone')


# Google cloud authentication ->
def init_gcloud(gcloud_path):
    out, _ = cmd(f'{gcloud_path} auth list')
    if not out:
        print('Please perform gcloud authentication to deploy Flow on GKE')
        cmd(f'{gcloud_path} auth login', std_output=True)


# List the projects and present it as options to user
def get_project(gcloud_path):
    project_list = []
    output, _ = cmd(f'{gcloud_path} projects list --format=json')
    projects = output.decode('utf-8')
    projects = json.loads(projects)
    for proj in projects:
        project_list.append(proj['projectId'])

    return ask_projects(project_list)


def get_region(gcloud_path):
    regions_list = []
    output, _ = cmd(f'{gcloud_path} compute regions list --format=json')
    regions = output.decode('utf-8')
    regions = json.loads(regions)
    for region in regions:
        regions_list.append(region['name'])

    return ask_regions(regions_list)


def get_zone(region, gcloud_path):
    zones_list = []
    output, _ = cmd(f'{gcloud_path} compute zones list --format=json')
    zones = output.decode('utf-8')
    zones = json.loads(zones)
    for zone in zones:
        if region in zone['name']:
            zones_list.append(zone['name'])

    return ask_zones(zones_list)


def final_confirmation():
    questions = [
        {
            'type': 'list',
            'name': 'proceed',
            'message': 'Creating a cluster will create some costs. '
            'Are you sure you want to continue? '
            'Prices can be checked here: '
            'https://cloud.google.com/kubernetes-engine/pricing',
            'choices': [
                {'name': '💸🔥 yes', 'value': True},
                {'name': ':⛔ no', 'value': False},
            ],
        }
    ]
    proceed = maybe_prompt_user(questions, 'proceed')
    if not proceed:
        cowsay.cow('see you soon 👋')
        exit(0)


def create_gke_cluster():
    gcloud_path, _ = cmd('which gcloud')
    gcloud_path = gcloud_path.strip()
    if not gcloud_path:
        gcloud_path = user('~/.cache/jina-now/google-cloud-sdk/bin/gcloud')
    else:
        gcloud_path = gcloud_path.decode('utf-8')
    application_name = 'jina-now'
    init_gcloud(gcloud_path)
    proj = get_project(gcloud_path)
    cmd(f'{gcloud_path} config set project {proj}')
    region = get_region(gcloud_path)
    cmd(f'{gcloud_path} config set compute/region {region}')
    zone = get_zone(region, gcloud_path)
    cmd(f'{gcloud_path} config set compute/zone {zone}')
    final_confirmation()
    out, _ = cmd(f'{gcloud_path} container clusters list')
    out = out.decode('utf-8')
    if application_name in out and zone in out:
        with yaspin(
            sigmap=sigmap, text='Cluster exists already', color='green'
        ) as spinner:
            spinner.ok('✅')
    else:
        with yaspin(
            custom_spinner().weather, sigmap=sigmap, text="Create cluster"
        ) as spinner:
            cmd(
                f'/bin/bash {cur_dir}/scripts/gke_deploy.sh {application_name} {gcloud_path}'
            )
            spinner.ok('🌥')


if __name__ == '__main__':
    create_gke_cluster()



import os
import pickle
import warnings
from os.path import join as osp

from yaspin import yaspin

from now.data_loading.data_loading import fill_missing, load_data
from now.deployment.flow import deploy_flow
from now.dialog import UserInput
from now.utils import sigmap


def save_mean(da, tmpdir):
    mean = da.embeddings.mean(0)
    if not os.path.exists(osp(tmpdir, 'now/hub/head_encoder/')):
        os.makedirs(osp(tmpdir, 'now/hub/head_encoder/'))
    with open(osp(tmpdir, 'now/hub/head_encoder/mean.bin'), 'wb') as f:
        pickle.dump(mean, f)


def is_finetuning(dataset_name, dataset):
    # finetuning for some datasets is deactivated since the generalization ability is lost
    # they can be reactivated once this ticket is done:
    # https://github.com/jina-ai/now/issues/76
    if dataset_name in [
        'tll',
        'nft-monkey',
        # 'deepfashion',
        'nih-chest-xrays',
        'geolocation-geoguessr',
        'stanford-cars',
        # 'bird-species',
        'best-artworks',
        'lyrics',
        'lyrics-10000',
        'rock-lyrics',
        'pop-lyrics',
        'rap-lyrics',
        'indie-lyrics',
        'metal-lyrics',
    ]:
        return False
    for d in dataset:
        if 'finetuner_label' in d.tags:
            return True
    return False


def run(user_input: UserInput, is_debug, tmpdir, kubectl_path: str):
    """
    Args:
        user_input: User input arguments
        is_debug: if True it also works on small datasets
    """
    # bring back at some point to make this configurable by the user
    (
        final_layer_output_dim,
        embedding_size,
        batch_size,
        train_val_split_ratio,
        num_default_val_queries,
    ) = parse_user_input(user_input.quality, is_debug)

    dataset, ds_type = load_data(
        user_input.output_modality,
        user_input.data,
        user_input.quality,
        user_input.is_custom_dataset,
        user_input.custom_dataset_type,
        user_input.dataset_secret,
        user_input.dataset_url,
        user_input.dataset_path,
    )

    finetuning = is_finetuning(user_input.data, dataset)

    if not finetuning:
        embedding_size = int(final_layer_output_dim / 2)
    dataset = {
        'index': dataset,
        'train': None,
        'val': None,
        'val_query': None,
        'val_index': None,
        'val_query_image': None,
        'val_index_image': None,
    }

    if finetuning:
        from now.finetuning.finetuning import add_clip_embeddings, finetune_layer
        from now.hub.head_encoder.head_encoder import extend_embeddings
        from now.hub.hub import push_to_hub
        from now.improvements.improvements import show_improvement

        add_clip_embeddings(
            dataset,
            user_input.model_variant,
            tmpdir,
            kubectl_path=kubectl_path,
        )
        extend_embeddings(dataset['index'], final_layer_output_dim)
        save_mean(dataset['index'], tmpdir)
        fill_missing(dataset, train_val_split_ratio, num_default_val_queries, is_debug)

        # if False:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            finetuned_model_path = finetune_layer(
                dataset, batch_size, final_layer_output_dim, embedding_size, tmpdir
            )

        with yaspin(sigmap=sigmap, text="Create overview", color="green") as spinner:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    show_improvement(
                        user_input.data,
                        user_input.quality,
                        dataset['val_query_image'],
                        dataset['val_index_image'],
                        dataset['val_query'],
                        dataset['val_index'],
                        final_layer_output_dim,
                        embedding_size,
                        finetuned_model_path,
                        class_label='finetuner_label',
                    )
            except Exception as e:
                pass
            spinner.ok('🖼')
        print(
            f'before-after comparison result is saved in the current working directory as image'
        )
        executor_name = push_to_hub(tmpdir)
    else:
        executor_name = None
    # executor_name = 'FineTunedLinearHeadEncoder:93ea59dbd1ee3fe0bdc44252c6e86a87/
    # linear_head_encoder_2022-02-20_20-35-15'
    # print('###executor_name', executor_name)
    # executor_name = 'FineTunedLinearHeadEncoder:93ea59dbd1ee3fe0bdc44252c6e86a87/
    # deleteme_2022-02-06_13-20-37'
    (
        gateway_host,
        gateway_port,
        gateway_host_internal,
        gateway_port_internal,
    ) = deploy_flow(
        executor_name=executor_name,
        output_modality=user_input.output_modality,
        index=dataset['index'],
        vision_model=user_input.model_variant,
        final_layer_output_dim=final_layer_output_dim,
        embedding_size=embedding_size,
        tmpdir=tmpdir,
        finetuning=finetuning,
        kubectl_path=kubectl_path,
    )
    return gateway_host, gateway_port, gateway_host_internal, gateway_port_internal


def parse_user_input(quality, is_debug):
    if is_debug:
        train_val_split_ratio = 0.5
        batch_size = 10
    else:
        batch_size = 128
        train_val_split_ratio = 0.9
    num_default_val_queries = 10

    if quality == 'ViT-L14':
        final_layer_output_dim = 768 * 2
    else:
        final_layer_output_dim = 512 * 2
    embedding_size = 128

    return (
        final_layer_output_dim,
        embedding_size,
        batch_size,
        train_val_split_ratio,
        num_default_val_queries,
    )



import base64
import os
import sys
from copy import deepcopy
from urllib.request import urlopen

import av
import numpy as np
import streamlit as st
from docarray import DocumentArray
from jina import Client, Document
from streamlit_webrtc import ClientSettings, webrtc_streamer

WEBRTC_CLIENT_SETTINGS = ClientSettings(
    rtc_configuration={"iceServers": [{"urls": ["stun:stun.l.google.com:19302"]}]},
    media_stream_constraints={"video": True, "audio": False},
)


root_data_dir = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets/'
)

ds_set = {
    'nft-monkey',
    'deepfashion',
    'nih-chest-xrays',
    'stanford-cars',
    'bird-species',
    'best-artworks',
    'geolocation-geoguessr',
    'rock-lyrics',
    'pop-lyrics',
    'rap-lyrics',
    'indie-lyrics',
    'metal-lyrics',
}


def deploy_streamlit():
    """
    We want to provide the end-to-end experience to the user.
    Please deploy a streamlit frontend on k8s/local to access the api.
    You can get the starting point for the streamlit application from alex.
    """
    setup_session_state()
    print('Run Streamlit with:', sys.argv)
    _, host, port, output_modality, data = sys.argv
    da_img = None
    da_txt = None

    # General
    TOP_K = 9
    DEBUG = os.getenv("DEBUG", False)
    DATA_DIR = "../data/images/"

    if data in ds_set:
        if output_modality == 'image':
            output_modality_dir = 'jpeg'
            data_dir = root_data_dir + output_modality_dir + '/'
            da_img, da_txt = load_data(data_dir + data + '.img10.bin'), load_data(
                data_dir + data + '.txt10.bin'
            )
        elif output_modality == 'text':
            # for now deactivated sample images for text
            output_modality_dir = 'text'
            data_dir = root_data_dir + output_modality_dir + '/'
            da_txt = load_data(data_dir + data + '.txt10.bin')

    class UI:
        about_block = """
        ### About
        This is a meme search engine using [Jina's neural search framework](https://github.com/jina-ai/jina/).
        - [Live demo](https://examples.jina.ai/memes)
        - [Play with it in a notebook](https://colab.research.google.com/github/jina-ai/workshops/blob/main/memes/meme_search.ipynb) (t-only)
        - [Repo](https://github.com/alexcg1/jina-meme-search)
        - [Dataset](https://www.kaggle.com/abhishtagatya/imgflipscraped-memes-caption-dataset)
        """

        css = """
        <style>
            .reportview-container .main .block-container{{
                max-width: 1200px;
                padding-top: 2rem;
                padding-right: 2rem;
                padding-left: 2rem;
                padding-bottom: 2rem;
            }}
            .reportview-container .main {{
                color: "#111";
                background-color: "#eee";
            }}
        </style>
        """

    def search_by_t(input, server, port, limit=TOP_K):
        print('initialize client at', server, port)
        client = Client(host=server, protocol="grpc", port=port)
        print('search text', server, port)
        response = client.search(
            Document(text=input),
            parameters={"limit": limit, 'filter': {}},
            return_results=True,
            show_progress=True,
        )

        return response[0].matches

    def search_by_file(document, server, port, limit=TOP_K):
        """
        Wrap file in Jina Document for searching, and do all necessary conversion to make similar to indexed Docs
        """
        print('connect client to ', server, port)
        client = Client(host=server, protocol="grpc", port=port)
        query_doc = document
        if query_doc.blob != b'':
            query_doc.convert_blob_to_image_tensor()
        query_doc.set_image_tensor_shape((224, 224))
        response = client.search(
            query_doc,
            parameters={"limit": limit, 'filter': {}},
            return_results=True,
            show_progress=True,
        )

        return response[0].matches

    def convert_file_to_document(query):
        data = query.read()
        doc = Document(blob=data)
        return doc

    # Layout
    st.set_page_config(page_title="NOW", page_icon='https://jina.ai/favicon.ico')

    st.markdown(
        body=UI.css,
        unsafe_allow_html=True,
    )
    col1, mid, col2 = st.columns([1, 1, 20])
    with col1:
        st.image('https://jina.ai/favicon.ico', width=60)
    with col2:
        st.header("NOW ")

    # design and create toggle button
    st.write(
        '<style>div.row-widget.stRadio > div{flex-direction:row;justify-content: center;} </style>',
        unsafe_allow_html=True,
    )
    st.write(
        '<style>div.st-bf{flex-direction:column;} div.st-ag{font-weight:bold;padding-right:50px;}</style>',
        unsafe_allow_html=True,
    )
    if output_modality == 'image':
        media_type = st.radio(
            '',
            ["Text", "Image", 'Webcam'],
            on_change=clear_match,
        )
    elif output_modality == 'text':
        media_type = st.radio(
            '',
            ["Image", "Text", 'Webcam'],
            on_change=clear_match,
        )

    if media_type == "Image":
        upload_c, preview_c = st.columns([12, 1])
        query = upload_c.file_uploader("")
        if query:
            doc = convert_file_to_document(query)
            st.image(doc.blob, width=160)
            st.session_state.matches = search_by_file(
                document=doc, server=host, port=port
            )
        if da_img is not None:
            st.subheader("samples:")
            img_cs = st.columns(5)
            txt_cs = st.columns(5)
            for doc, c, txt in zip(da_img, img_cs, txt_cs):
                with c:
                    st.image(doc.blob if doc.blob else doc.tensor, width=100)
                with txt:
                    if st.button('Search', key=doc.id):
                        st.session_state.matches = search_by_file(
                            document=doc,
                            server=host,
                            port=port,
                        )

    elif media_type == "Text":
        query = st.text_input("", key="text_search_box")
        if query:
            st.session_state.matches = search_by_t(input=query, server=host, port=port)
        if st.button("Search", key="text_search"):
            st.session_state.matches = search_by_t(input=query, server=host, port=port)
        if da_txt is not None:
            st.subheader("samples:")
            c1, c2, c3 = st.columns(3)
            c4, c5, c6 = st.columns(3)
            for doc, col in zip(da_txt, [c1, c2, c3, c4, c5, c6]):
                with col:
                    if st.button(doc.content, key=doc.id, on_click=clear_text):
                        st.session_state.matches = search_by_t(
                            input=doc.content, server=host, port=port
                        )

    elif media_type == 'Webcam':
        snapshot = st.button('Snapshot')

        class VideoProcessor:
            snapshot: np.ndarray = None

            def recv(self, frame):
                self.snapshot = frame.to_ndarray(format="rgb24")
                return av.VideoFrame.from_ndarray(self.snapshot, format='rgb24')

        ctx = webrtc_streamer(
            key="jina-now",
            video_processor_factory=VideoProcessor,
            client_settings=WEBRTC_CLIENT_SETTINGS,
        )

        if ctx.video_processor:
            if snapshot:
                query = ctx.video_processor.snapshot
                st.image(query, width=160)
                st.session_state.snap = query
                doc = Document(tensor=query)
                doc.convert_image_tensor_to_blob()
                st.session_state.matches = search_by_file(
                    document=doc, server=host, port=port
                )
            elif st.session_state.snap is not None:
                st.image(st.session_state.snap, width=160)
        else:
            clear_match()

    if st.session_state.matches:
        matches = deepcopy(st.session_state.matches)
        st.header('Search results')
        # Results area
        c1, c2, c3 = st.columns(3)
        c4, c5, c6 = st.columns(3)
        c7, c8, c9 = st.columns(3)
        all_cs = [c1, c2, c3, c4, c5, c6, c7, c8, c9]
        # # TODO dirty hack to filter out text. Instead output modality should be passed as parameter
        # matches = [m for m in matches if m.tensor is None]
        for m in matches:
            m.scores['cosine'].value = 1 - m.scores['cosine'].value
        sorted(matches, key=lambda m: m.scores['cosine'].value, reverse=True)
        matches = [
            m
            for m in matches
            if m.scores['cosine'].value > st.session_state.min_confidence
        ]
        for c, match in zip(all_cs, matches):
            match.mime_type = output_modality

            if output_modality == 'text':
                display_text = match.text
                body = f"<!DOCTYPE html><html><body><blockquote>{display_text}</blockquote>"
                if match.tags.get('additional_info'):
                    body += (
                        f"<figcaption>{match.tags.get('additional_info')}</figcaption>"
                    )
                body += "</body></html>"
                c.markdown(
                    body=body,
                    unsafe_allow_html=True,
                )
            elif match.uri is not None:
                if match.blob != b'':
                    match.convert_blob_to_datauri()
                if match.tensor is not None:
                    match.convert_image_tensor_to_uri()
                c.image(match.convert_blob_to_datauri().uri)
        st.markdown("""---""")
        st.session_state.min_confidence = st.slider(
            'Confidence threshold',
            0.0,
            1.0,
            key='slider',
            on_change=update_conf,
        )


def update_conf():
    st.session_state.min_confidence = st.session_state.slider


def clear_match():
    st.session_state.matches = None
    st.session_state.slider = 0.0
    st.session_state.min_confidence = 0.0
    st.session_state.snap = None


def clear_text():
    st.session_state.text_search_box = ''


def load_data(data_path: str) -> DocumentArray:
    if data_path.startswith('http'):
        try:
            # TODO try except is used as workaround
            # in case load_data is called two times from two frontends it can happen that
            # one of the calls created the directory right after checking that it does not exist
            # this caused errors. Now the error will be ignored.
            # Can not use `exist=True` because it is not available in py3.7
            os.makedirs('data/tmp')
        except:
            pass
        url = data_path
        data_path = (
            f"data/tmp/{base64.b64encode(bytes(url, 'utf-8')).decode('utf-8')}.bin"
        )
        if not os.path.exists(data_path):
            with urlopen(url) as f:
                content = f.read()
            with open(data_path, 'wb') as f:
                f.write(content)

    try:
        da = DocumentArray.load_binary(data_path)
    except Exception:
        da = DocumentArray.load_binary(data_path, compress='gzip')
    return da


def setup_session_state():
    if 'matches' not in st.session_state:
        st.session_state.matches = None

    if 'min_confidence' not in st.session_state:
        st.session_state.min_confidence = 0.0

    if 'im' not in st.session_state:
        st.session_state.im = None

    if 'snap' not in st.session_state:
        st.session_state.snap = None


if __name__ == '__main__':
    deploy_streamlit()
